use std::{iter::Peekable, vec::IntoIter};

use serde::Serialize;
use sqlx::{postgres::PgRow, Pool, Postgres, Row};
use uuid::Uuid;

#[derive(Clone)]
pub struct DB {
    connection_pool: Pool::<Postgres>
}

#[derive(Serialize)]
pub struct Site {
    uuid: Uuid,
    title: String,
    base_url: String,
    theme: String,
    boards: Vec<Board>
}

#[derive(Serialize, Clone)]
pub struct Board {
    uuid: Uuid,
    title: String,
    description: String,
    threads: Vec<Thread>
}

#[derive(Serialize, Clone)]
pub struct Thread {
    uuid: Uuid,
    title: String,
    posts: Vec<Post>
}

#[derive(Serialize, Clone)]
pub struct Post {
    uuid: Uuid,
    contents: String,
    author: User
}

#[derive(Serialize, Clone)]
pub struct User {
    uuid: Uuid,
    email: String,
    username: String,
    password_hash: String,
    is_admin: bool
}

impl DB {
    pub async fn init() -> Self {
        let db_url = std::env::var("DATABASE_URL").expect("Please provide DATABASE_URL in environment");
        let pool = Pool::<Postgres>::connect(db_url.as_str()).await.expect("Could not connect to database");

        sqlx::migrate!()
            .run(&pool)
            .await
            .expect("Could not run database migrations");

        DB {
            connection_pool: pool
        }
    }

    pub async fn get_site_data(&self) -> Result<Site, String> {
        let rows : Vec<PgRow> = sqlx::query(
            r#"select
                site_uuid,
                site_title,
                site_base_url,
                site_theme,
                board_uuid,
                board_title,
                board_description,
                thread_uuid,
                thread_title,
                post_uuid as "post_uuid?",
                post_contents as "post_contents?",
                user_uuid,
                user_email,
                user_username,
                user_password_hash,
                user_is_admin
            from forum.site
                left join forum.board on board_site = site_uuid
                left join forum.thread on thread_board = board_uuid
                left join forum.post on post_thread = thread_uuid
                left join forum.user on post_author = user_uuid
            "#).fetch_all(&self.connection_pool).await.expect("Could not connect to database");

        let mut rows_iter = rows.into_iter().peekable();

        let site = Site::parse_from_rows(&mut rows_iter, "site_".to_string());

        if let Some(site) = site {
            Ok(site)
        } else {
            Err("Could not find site in DB".to_string())
        }
    }
}

type Rows = Peekable<IntoIter<PgRow>>;
trait ParseFromRows {
    fn parse_from_rows<'a>(rows: &'a mut Rows, item_prefix: String) -> Option<Self> where Self: Sized;
 }

macro_rules! parse_field {
    ( $row_name:ident, $item_prefix:ident, $field_name:literal ) => {
        $row_name.get(format!("{}{}", $item_prefix, $field_name).as_str())
    }
}

macro_rules! parse_array {
    ( $rows:ident, $current_uuid:ident, $uuid_field:ident, $array_type:ty, $field_prefix:literal) => {
        {
            let mut items = Vec::new();

            loop {
                if let Some(item) = <$array_type>::parse_from_rows($rows, $field_prefix.to_string()) {
                    items.push(item);
                }

                $rows.next();
                let next = $rows.peek();

                if next.is_none() {
                    break
                }

                let uuid_result = next.unwrap().try_get::<Option<Uuid>, &str>(&$uuid_field.as_str());

                if uuid_result.is_err() || uuid_result.as_ref().unwrap().is_none() {
                    break
                }

                let next_uuid = uuid_result.as_ref().unwrap().unwrap().clone();

                if next_uuid != $current_uuid {
                    break
                }
            }

            items
        }
    };
}

impl ParseFromRows for User {
    fn parse_from_rows(rows: &mut Rows, item_prefix: String) -> Option<Self> {
        let next = rows.peek();
        if next.is_none() {
            return None
        }

        let next = next.unwrap();
        let uuid_field = format!("{}uuid", item_prefix) ;
        let uuid_result = next.try_get::<Option<Uuid>, &str>(&uuid_field.as_str());

        if uuid_result.is_err() || uuid_result.as_ref().unwrap().is_none() {
            return None
        }

        let uuid = uuid_result.as_ref().unwrap().unwrap().clone();
        let email = parse_field!(next, item_prefix, "email");
        let username = parse_field!(next, item_prefix, "username");
        let password_hash = parse_field!(next, item_prefix, "password_hash");
        let is_admin = parse_field!(next, item_prefix, "is_admin");

        Some(Self {
            uuid,
            email,
            username,
            password_hash,
            is_admin
        })
    }
}

impl ParseFromRows for Post {
    fn parse_from_rows(rows: &mut Rows, item_prefix: String) -> Option<Self> {
        let next = rows.peek();
        if next.is_none() {
            return None
        }

        let next = next.unwrap();
        let uuid_field = format!("{}uuid", item_prefix) ;
        let uuid_result = next.try_get::<Option<Uuid>, &str>(&uuid_field.as_str());

        if uuid_result.is_err() || uuid_result.as_ref().unwrap().is_none() {
            return None
        }

        let uuid = uuid_result.as_ref().unwrap().unwrap().clone();
        let contents = parse_field!(next, item_prefix, "contents");
        let author = User::parse_from_rows(rows, "user_".to_string()).unwrap();

        Some(Self {
            uuid,
            author,
            contents
        })
    }
}

impl ParseFromRows for Thread {
    fn parse_from_rows(rows: &mut Rows, item_prefix: String) -> Option<Self> {
        let next = rows.peek();
        if next.is_none() {
            return None
        }

        let next = next.unwrap();
        let uuid_field = format!("{}uuid", item_prefix) ;
        let uuid_result = next.try_get::<Option<Uuid>, &str>(&uuid_field.as_str());

        if uuid_result.is_err() || uuid_result.as_ref().unwrap().is_none() {
            return None
        }

        let uuid = uuid_result.as_ref().unwrap().unwrap().clone();
        let title = parse_field!(next, item_prefix, "title");
        let posts = parse_array!(rows, uuid, uuid_field, Post, "post_");

        Some(Self {
            uuid,
            title,
            posts
        })
    }
}

impl ParseFromRows for Board {
    fn parse_from_rows(rows: &mut Rows, item_prefix: String) -> Option<Self> {
        let next = rows.peek();
        if next.is_none() {
            return None
        }

        let next = next.unwrap();
        let uuid_field = format!("{}uuid", item_prefix) ;
        let uuid_result = next.try_get::<Option<Uuid>, &str>(&uuid_field.as_str());

        if uuid_result.is_err() || uuid_result.as_ref().unwrap().is_none() {
            return None
        }

        let uuid = uuid_result.as_ref().unwrap().unwrap().clone();
        let title = parse_field!(next, item_prefix, "title");
        let description = parse_field!(next, item_prefix, "description");
        let threads = parse_array!(rows, uuid, uuid_field, Thread, "thread_");

        Some(Self {
            uuid,
            title,
            description,
            threads
        })
    }
}

impl ParseFromRows for Site {
    fn parse_from_rows(rows: &mut Rows, item_prefix: String) -> Option<Self> {
        let next = rows.peek();
        if next.is_none() {
            return None
        }

        let next = next.unwrap();
        let uuid_field = format!("{}uuid", item_prefix) ;
        let uuid_result = next.try_get::<Option<Uuid>, &str>(&uuid_field.as_str());

        if uuid_result.is_err() || uuid_result.as_ref().unwrap().is_none() {
            return None
        }

        let uuid = uuid_result.as_ref().unwrap().unwrap().clone();
        let title = parse_field!(next, item_prefix, "title");
        let base_url = parse_field!(next, item_prefix, "base_url");
        let theme = parse_field!(next, item_prefix, "theme");
        let boards = parse_array!(rows, uuid, uuid_field, Board, "board_");

        Some(Self {
            uuid,
            title,
            base_url,
            theme,
            boards
        })
    }
}
